"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const NodeUtils_1 = require("./NodeUtils");
const three_1 = require("three");
let _nodeId = 0;
class Node {
    constructor(nodeType = null) {
        this.nodeType = nodeType;
        this.updateType = constants_1.NodeUpdateType.None;
        this.uuid = three_1.MathUtils.generateUUID();
        Object.defineProperty(this, 'id', { value: _nodeId++ });
    }
    get type() {
        return this.constructor.name;
    }
    getHash( /*builder*/) {
        return this.uuid;
    }
    getUpdateType( /*builder*/) {
        return this.updateType;
    }
    getNodeType( /*builder*/) {
        return this.nodeType;
    }
    getReference(builder) {
        const hash = this.getHash(builder);
        const nodeFromHash = builder.getNodeFromHash(hash);
        return nodeFromHash || this;
    }
    update( /*frame*/) {
        console.warn('Abstract function.');
    }
    generate( /*builder, output*/) {
        console.warn('Abstract function.');
    }
    analyze(builder) {
        const refNode = this.getReference(builder);
        if (this !== refNode) {
            return refNode.analyze(builder);
        }
        const nodeData = builder.getDataFromNode(this);
        nodeData.dependenciesCount = nodeData.dependenciesCount === undefined ? 1 : nodeData.dependenciesCount + 1;
        const nodeKeys = NodeUtils_1.getNodesKeys(this);
        for (const property of nodeKeys) {
            this[property].analyze(builder);
        }
    }
    build(builder, output = null) {
        const refNode = this.getReference(builder);
        if (this !== refNode) {
            return refNode.build(builder, output);
        }
        builder.addNode(this);
        builder.addStack(this);
        const nodeData = builder.getDataFromNode(this);
        const isGenerateOnce = this.generate.length === 1;
        let snippet = null;
        if (isGenerateOnce) {
            const type = this.getNodeType(builder);
            snippet = nodeData.snippet;
            if (snippet === undefined) {
                snippet = this.generate(builder) || '';
                nodeData.snippet = snippet;
            }
            snippet = builder.format(snippet, type, output);
        }
        else {
            snippet = this.generate(builder, output) || '';
        }
        builder.removeStack(this);
        return snippet;
    }
    serialize(json) {
        const nodeKeys = NodeUtils_1.getNodesKeys(this);
        if (nodeKeys.length > 0) {
            const inputNodes = {};
            for (const property of nodeKeys) {
                inputNodes[property] = this[property].toJSON(json.meta).uuid;
            }
            json.inputNodes = inputNodes;
        }
    }
    deserialize(json) {
        if (json.inputNodes !== undefined) {
            const nodes = json.meta.nodes;
            for (const property in json.inputNodes) {
                const uuid = json.inputNodes[property];
                this[property] = nodes[uuid];
            }
        }
    }
    toJSON(meta) {
        const { uuid, type } = this;
        const isRoot = meta === undefined || typeof meta === 'string';
        if (isRoot) {
            meta = {
                textures: {},
                images: {},
                nodes: {},
            };
        }
        // serialize
        let data = meta.nodes[uuid];
        if (data === undefined) {
            data = {
                uuid,
                type,
                meta,
                metadata: {
                    version: 4.5,
                    type: 'Node',
                    generator: 'Node.toJSON',
                },
            };
            meta.nodes[data.uuid] = data;
            this.serialize(data);
            delete data.meta;
        }
        // TODO: Copied from Object3D.toJSON
        function extractFromCache(cache) {
            const values = [];
            for (const key in cache) {
                const data = cache[key];
                delete data.metadata;
                values.push(data);
            }
            return values;
        }
        if (isRoot) {
            const textures = extractFromCache(meta.textures);
            const images = extractFromCache(meta.images);
            const nodes = extractFromCache(meta.nodes);
            if (textures.length > 0)
                data.textures = textures;
            if (images.length > 0)
                data.images = images;
            if (nodes.length > 0)
                data.nodes = nodes;
        }
        return data;
    }
}
Node.prototype.isNode = true;
exports.default = Node;
